"""
This file stores the package version.
"""

# SPDX-FileCopyrightText: 2023-present Remco <remco@educationwarehouse.nl>
#
# SPDX-License-Identifier: MIT

# __version__ = "0.10.4"
__version__ = "1.0.0a16"



import pytest
from testcontainers.redis import RedisContainer

from src.edwh_migrate import activate_migrations, get_config

from .fixtures import *


@pytest.fixture
def redis(request):
    r = RedisContainer("redis:6.2.6")
    request.addfinalizer(r.stop)

    r.start()
    return r


@pytest.fixture
def valkey(request):
    r = RedisContainer("valkey/valkey:7.2.5")
    request.addfinalizer(r.stop)

    r.start()
    return r


def test_redis_cleared(tmp_just_implemented_features_sqlite_db_file: Path, redis: RedisContainer):
    config = get_config()

    redis_host = redis.get_container_host_ip()
    redis_port = redis.get_exposed_port(redis.port)
    config.redis_host = f"{redis_host}:{redis_port}"

    redis_client = redis.get_client()

    redis_client.set("a", 1)
    assert redis_client.get("a")
    assert not redis_client.get("b")

    activate_migrations(config)

    assert not redis_client.get("a")
    assert not redis_client.get("b")


def test_valkey_cleared(tmp_just_implemented_features_sqlite_db_file: Path, valkey: RedisContainer):
    config = get_config()

    redis_host = valkey.get_container_host_ip()
    redis_port = valkey.get_exposed_port(valkey.port)
    config.redis_host = f"{redis_host}:{redis_port}"

    redis_client = valkey.get_client()

    redis_client.set("a", 1)
    assert redis_client.get("a")
    assert not redis_client.get("b")

    activate_migrations(config)

    assert not redis_client.get("a")
    assert not redis_client.get("b")



import pytest

from src.edwh_migrate import activate_migrations, get_config, list_migrations, migration, setup_db

from .fixtures import clean_migrate, sqlite_empty, tmp_empty_sqlite_db_file, tmp_sqlite_folder  # noqa


def test_cascading_order(tmp_empty_sqlite_db_file, clean_migrate):
    # file 1
    @migration()
    def c1(db):
        return True

    @migration()
    def c2(db):
        return True

    @migration()
    def c3(db):
        return True

    @migration()
    def c4(db):
        return True

    @migration()
    def c5(db):
        return True

    @migration()
    def c6(db):
        return True

    # file 2
    @migration(requires=[c2])
    def a1(db):
        return True

    @migration()
    def a2(db):
        return True

    @migration(requires=[c4, "c6"])
    def a3(db):
        return True

    @migration()
    def a4(db):
        return True

    @migration()
    def a5(db):
        return True

    config = get_config()

    expected_order = ["c1", "c2", "a1", "c3", "c4", "c5", "c6", "a3", "a2", "a4", "a5"]

    assert list(list_migrations(config).keys()) == expected_order

    assert activate_migrations(config)

    db = setup_db(config=config)

    assert db(db.ewh_implemented_features).select("name").column("name") == expected_order


def test_cascade_on_nonexisting(tmp_empty_sqlite_db_file, clean_migrate):
    @migration()
    def very_real(db): ...

    try:

        @migration(requires=["very_real", "very_fake"])
        def depends_on(db): ...

        assert False, "unreachable"
    except ValueError as e:
        assert "very_fake" in str(e)


def test_duplicate(tmp_empty_sqlite_db_file, clean_migrate):
    @migration()
    def very_first_001(db):
        return False

    with pytest.raises(ValueError):

        @migration()
        def very_first_001(db):
            return True



try:
    import psycopg2
    import psycopg2.errors

    PostgresError = psycopg2.errors.Error

    PostgresUndefinedTable = psycopg2.errors.UndefinedTable

except ImportError:  # pragma: no cover
    """
    Create fallback exceptions for when psycopg2 is not installed (e.g. when using `migrate` with sqlite only).
    """

    class PostgresError(Exception):
        """
        Placeholder for when psycopg2 is not installed.
        """

    class PostgresUndefinedTable(Exception):
        """
        Placeholder for when psycopg2 is not installed.
        """



import os
import tempfile
import textwrap
from contextlib import chdir

import pytest
from configuraptor import Singleton

from src.edwh_migrate import migrate
from src.edwh_migrate.migrate import Config, import_migrations, list_migrations


@pytest.fixture
def migrations_at_temp():
    with tempfile.TemporaryDirectory() as d:
        fname = f"{d}/migrations.py"
        with open(fname, "w") as f:
            f.write(
                textwrap.dedent(
                    """
            from src.edwh_migrate import migration
            
            @migration()
            def test(db): return True

            """
                )
            )
        yield d


@pytest.fixture()
def empty_temp():
    with tempfile.TemporaryDirectory() as d:
        yield d


@pytest.fixture()
def empty_config():
    migrate.migrations.reset()
    Singleton.clear()

    return Config.load({})


def test_list(migrations_at_temp: str, empty_config: Config):
    with chdir(migrations_at_temp):
        found = list_migrations(empty_config)
        assert list(found.keys()) == ["test"]


def test_args(migrations_at_temp: str, empty_config: Config):
    assert import_migrations([migrations_at_temp], empty_config)
    migrate.migrations.reset()

    assert import_migrations([f"{migrations_at_temp}/migrations.py"], empty_config)
    migrate.migrations.reset()

    assert not import_migrations([f"{migrations_at_temp}/no_migrations.py"], empty_config)
    migrate.migrations.reset()

    assert not import_migrations([f"/no_tmp"], empty_config)
    migrate.migrations.reset()


def test_config(migrations_at_temp: str, empty_config: Config):
    empty_config.migrations_file = f"{migrations_at_temp}/migrations.py"
    assert import_migrations([], empty_config)


def test_fail(migrations_at_temp: str, empty_temp: str, empty_config: Config):
    with chdir(empty_temp):
        assert not import_migrations([], empty_config)


def test_local(migrations_at_temp: str, empty_config: Config):
    empty_config.migrations_file = "/fake/file"
    with chdir(migrations_at_temp):
        assert import_migrations([], empty_config)



import abc
import os
import types
import typing
import warnings

from pydal import DAL

from .constants import CURRENT_MIGRATION
from .helpers import classproperty

E = typing.TypeVar("E", bound=Exception)


class ViewMigrationManager(abc.ABC):
    """
    A base class for managing the lifecycle of view migrations in a database using context management.

    This class provides an abstract framework for creating, managing, and reversing database views,
    ensuring that migrations are properly handled with dependencies between different migrations.

    Attributes:
        uses (list[typing.Type["ViewMigrationManager"]]): A list of other ViewMigrationManager classes
            that the current class depends on.

    Example:
        1. Define a subclass that implements specific migration logic:

        ```python
        class MyExampleView_V2(ViewMigrationManager):
            def up(self):
                self.db.executesql(
                    '''
                    CREATE MATERIALIZED VIEW my_example_view AS
                    SELECT id, name FROM my_table;
                    '''
                )

            def down(self):
                self.db.executesql(
                    '''
                    DROP MATERIALIZED VIEW IF EXISTS my_example_view;
                    '''
                )
        ```

        2. Use the subclass in a migration function:

        ```python
        @migration
        def upgrade_some_source_table_that_my_example_view_depends_on(db):
            with MyExampleView_V2(db):
                db.executesql('''
                -- Perform operations that might affect `my_example_view`
                ''')
            db.commit()
            return True
        ```

        In this example:
        - The `down` method will drop the `my_example_view` before the block of code is executed.
        - The SQL operations inside the block will run.
        - The `up` method will recreate the `my_example_view` after the block completes.

    Methods:
        __init__(db: DAL):
            Initializes the manager with the provided database connection.

        __init_subclass__():
            Handles subclass initialization, ensuring proper tracking of dependencies
            using the `uses` and `used_by` attributes.

        up():
            Abstract method to define the migration logic. Subclasses must implement this to
            define how the view should be created or modified.

        down():
            Abstract method to define the rollback logic. Subclasses must implement this to
            define how the view should be dropped or reverted.

        __enter__():
            Context management for reversing the migration before executing a block of code.

        __exit__(exc_type, exc_value, traceback):
            Context management for applying the migration after a block of code has executed,
            regardless of whether an exception was raised.
    """

    @abc.abstractmethod
    @classproperty
    def uses(cls) -> typing.Iterable[typing.Type["ViewMigrationManager"]]:
        """
        List/tuple of view migration classes that this one depends on.
        """
        warnings.warn(f"Class {cls} has default 'uses'? This may indicate missing dependencies!")
        return ()

    @classproperty
    def since(cls) -> str:
        """
        Migration that must have run before this migration becomes relevant.

        Otherwise, old migrations may break due to new dependencies and we definitely don't want that!
        """
        return ""

    @classproperty
    def until(cls) -> str:
        """
        Migration after which this class should no longer be used.

        This attribute specifies the migration after which this view migration manager
        should not be utilized. It ensures that new migrations do not use views or logic
        from this class beyond the specified migration.
        """
        return ""

    # note: manually setting `used_by` is deprecated!
    _used_by: list[typing.Type["ViewMigrationManager"]]

    may_go_up = False

    def __init__(self, db: DAL, cache: typing.Optional[dict] = None):
        """
        Initialize the ViewMigrationManager with a database connection.

        Args:
            db (DAL): The database connection object.
        """
        if cache is None:
            cache = {}

        self.db = db

        used_by = getattr(self, "_used_by", [])

        assert self.__class__ not in used_by, "Recursion prevented..."

        # self.instances = tuple(_(db, cache) for _ in used_by)

        instances = []
        for dep in used_by:
            if dep in cache:
                # re-used classes should use the same instance (but configuraptor.Singleton doesn't work here):
                i = cache[dep]
            else:
                i = dep(db, cache)
                cache[dep] = i
            instances.append(i)

        self.instances = instances

    def __init_subclass__(cls) -> None:
        if not hasattr(cls, "_used_by"):
            # note: this has to be created here,
            # otherwise 'used_by' is a shared reference between all subclasses!!!
            cls._used_by = []

        # pycharm doesn't really understand abstract class properties so cast the type here:
        dependencies = typing.cast(typing.Iterable[typing.Type["ViewMigrationManager"]], cls.uses)

        for dependency_cls in dependencies:
            dependency_cls._used_by.append(cls)

    def should_run(self) -> bool:
        """
        Perform checks to know whether this dependency is active.
        """
        return self.check_since() and self.check_until()

    def check_since(self) -> bool:
        """
        If a 'since' is specified, that migration should have run already at this point.
        """
        db = self.db

        if not (since := self.since):
            return True

        if os.environ[CURRENT_MIGRATION] == since:
            return True

        query = db.ewh_implemented_features.name == since
        query &= db.ewh_implemented_features.installed == True

        return db(query).count() > 0

    def check_until(self) -> bool:
        """
        If an 'until' is specified, that migration should not have run yet at this point.
        """
        db = self.db

        if not (until := self.until):
            return True

        query = db.ewh_implemented_features.name == until
        query &= db.ewh_implemented_features.installed == True

        return db(query).count() == 0

    @abc.abstractmethod
    def up(self) -> None:
        """
        Defines the logic to apply the migration, such as creating or modifying views.
        This method should be implemented in subclasses for the specific migration task.
        """

    @abc.abstractmethod
    def down(self) -> None:
        """
        Defines the logic to reverse the migration, such as dropping or reverting views.
        This method should be implemented in subclasses for the specific migration task.
        """

    def __enter__(self) -> None:
        """
        Context management method for entering the runtime context related to the migration.
        By default, this calls the `down` method to reverse or remove the migration before executing
        the block of code.

        Returns:
            ViewMigrationManager: The current instance of the migration manager.
        """
        if not self.should_run():
            return

        for item in reversed(self.instances):
            item.__enter__()

        self.down()
        self.may_go_up = True

    def __exit__(self, exc_type: typing.Type[E], exc_value: E, tb: types.TracebackType) -> None:
        """
        Context management method for exiting the runtime context related to the migration.
        This method calls the `up` method to apply the migration after the block of code finishes,
        regardless of whether an exception was raised.

        Args:
            exc_type (type): The exception type raised during execution (if any).
            exc_value (Exception): The exception instance raised during execution (if any).
            tb (traceback): The traceback object related to the exception (if any).
        """
        if exc_type:
            # block failed, don't try to go up!
            return

        if not self.should_run():
            return

        if not self.may_go_up:
            # didn't go down properly or already executed!
            return

        self.up()
        self.may_go_up = False

        for item in self.instances:
            item.__exit__(exc_type, exc_value, tb)
